/****************************************************************************
**  Cube callgraph plugin                                                  **
*****************************************************************************
**  Copyright (c) 2014                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/

#include "config.h"

#include <QVBoxLayout>
#include <QtPlugin>
#include <QTemporaryFile>
#include "CallgraphPlugin.h"
#include "SvgView.h"
#include "PluginServices.h"

/**  If the plugin doesn't load, the most likely reason is an undefined reference.
     To check this, call "make check-libexample-plugin" in the build-backend directory.
 */

using namespace callgraph_plugin;

#if QT_VERSION < 0x050000
Q_EXPORT_PLUGIN2( CallGraph, CallgraphPlugin ); // ( PluginName, ClassName )
#endif

/**
 * initializes the plugin, called if plugin is started
 */
bool
CallgraphPlugin::cubeOpened( cubepluginapi::PluginServices* service )
{
    this->service = service;
    this->cube = service->getCube();

    connect( service, SIGNAL( contextMenuIsShown( cubepluginapi::TreeType, cubepluginapi::TreeItem * ) ),
             this, SLOT( contextMenuIsShown( cubepluginapi::TreeType, cubepluginapi::TreeItem * ) ) );

    QMenu* menu = service->enablePluginMenu();
    addMenu( menu );
   
    return true; // initialisation is ok => plugin should be shown
}

/**
 * @brief CallgraphPlugin::contextMenuIsShown is called, if the user clicks with right mouse button on a tree
 * item.
 */
void
CallgraphPlugin::contextMenuIsShown( cubepluginapi::TreeType type, cubepluginapi::TreeItem* )
{
  if ( type == cubepluginapi::METRICTREE )
  {
    QAction* contextAction = service->addContextMenuItem( type,(service->getSelections(cubepluginapi::CALLTREE  ).size() > 1)? "Generate graphs":"Generate graph" );
    connect( contextAction, SIGNAL( triggered() ), this, SLOT( generateDotFileTop() ) );
  }
if ( type == cubepluginapi::CALLTREE )
  {
     QAction* contextAction = service->addContextMenuItem( type, (service->getSelections(cubepluginapi::CALLTREE  ).size() > 1)? "Generate graphs":"Generate graph" );
        
    connect( contextAction, SIGNAL( triggered() ), this, SLOT( generateDotFileSelected() ) );
  }
  
  
}



void
CallgraphPlugin::generateDotFile( const QList<cubepluginapi::TreeItem *>& items)
{
    QString filter = inputGroup->checkedAction()->text(); // selected graphviz tool
    QString output = outputGroup->checkedAction()->text(); // selected output format
    dotContents.clear();
    
    foreach (cubepluginapi::TreeItem *item, items)
    {
        fillDotHash( item );
    }

    // generate input dot file
    QTemporaryFile dotfile;
    //QFile file(dotFileName);
    if (dotfile.open())
    {
        QTextStream out(&dotfile);
        out << "digraph G {" << endl;
        foreach ( QString key, dotContents.keys() )
        {
            out << key <<  " [label=\"" << dotContents.value(key, 0.0) << "\"];" << endl;
        }
        out << "}" << endl;
    }
    dotfile.close();

    // get output file name for graphviz
    QString outputFileName = service->getCubeBaseName().append(".").append(output);
    QTemporaryFile tmpFile;
    //------------------------
    if ( output == "display" )
    {
        tmpFile.open();
        outputFileName = tmpFile.fileName();
    }

    // get output option for graphviz
    QString outputOption;
    if ( output == "display" )
    {
        outputOption = "-Tsvg";
    }
    else
    {
        outputOption= QString("-T%1").arg(output);
    }

    // execute selected graphviz tool
    QProcess process;
    process.start(filter, QStringList() << dotfile.fileName() << outputOption << "-o" << outputFileName);
    bool ok = process.waitForFinished();

    if (ok)
    {
        if ( output == "display" )
        {
            SvgView *svgView = new SvgView();
            svgView->openFile( tmpFile );
            svgView->setGeometry(100, 100, 600, 400);
            svgView->show();
        }
        else {
            qDebug() << "output saved to " << outputFileName;
        }
    }
    else
    {
        qDebug() << "error starting dot" << process.exitCode();
    }
    tmpFile.close();
}



void
CallgraphPlugin::generateDotFileSelected()
{

    const QList<cubepluginapi::TreeItem *> items = service->getSelections( cubepluginapi::CALLTREE );
    generateDotFile( items );
  
}

void
CallgraphPlugin::generateDotFileTop()
{
    const QList<cubepluginapi::TreeItem *> items = service->getTopLevelItems( cubepluginapi::CALLTREE );
    generateDotFile( items );
}




double
CallgraphPlugin::computeValue( cubepluginapi::TreeItem *metricItem, cubepluginapi::TreeItem* callItem,
                               cube::CalculationFlavour callState )
{
    double value = 0.;

    // set metric value data
    cube::list_of_metrics mv;
    cube::metric_pair mp;
    mp.first  = ( cube::Metric* ) metricItem->getCubeObject();
    mp.second =  metricItem->isExpanded() ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
    mv.push_back( mp );

    // calculation of value for call tree item, if metric tree is left of call tree
    value = cube->get_sev( mv, ( cube::Cnode* ) callItem->getCubeObject(), callState );

    return value;
}

/**
 * @brief CallgraphPlugin::fillDotHash
 * each key of the hash has the form "parent->child", the value is the exclusive value of the child for the selected metric
 * @param root call tree item, whose data is inserted into the dotContents hash
 */
void
CallgraphPlugin::fillDotHash(cubepluginapi::TreeItem *root)
{
    
    bool omit = (omitGroup->checkedAction()->text().compare(QString::fromStdString("yes") ) == 0);
    
    cubepluginapi::TreeItem *metricItem = service->getSelection( cubepluginapi::METRICTREE );

    foreach ( cubepluginapi::TreeItem *child, root->getChildren() )
    {
       QString key = QString( "\"" ).append( root->getName() ).append( "\" -> \"" ).append( child->getName() ).append( "\"" );

       double prev = dotContents.value( key, 0 );

       double current = computeValue(metricItem, child, cube::CUBE_CALCULATE_EXCLUSIVE );
       if ( (!omit) || (prev + current != 0))
            dotContents.insert( key, prev + current );

       fillDotHash( child );
    }
}


void
CallgraphPlugin::addMenu( QMenu * menu)
{
     
    QStringList filters = QStringList() 
#ifdef HAVE_DOT
    << "dot" 
#endif
#ifdef HAVE_NEATO  
    << "neato" 
#endif
#ifdef HAVE_TWOPI  
    << "twopi" 
#endif
#ifdef HAVE_CIRCO
    << "circo" 
#endif
#ifdef HAVE_FDP
    << "fdp" 
#endif
#ifdef HAVE_SFDP
    << "sfdp"
#endif
    
    ;
    QMenu *inputMenu = menu->addMenu("Select filter");
    inputGroup = new QActionGroup(inputMenu);
    foreach (QString filter, filters) {
        QAction *action = new QAction(filter, inputMenu);
        action->setCheckable(true);
        inputGroup->addAction( action );
        inputMenu->addAction( action );
    };
    inputGroup->actions().first()->setChecked( true );

    QStringList output = QStringList() << "display" << "svg" << "ps" << "pdf";
    QMenu *outputMenu = menu->addMenu("Select output");
    outputGroup = new QActionGroup(outputMenu);
    foreach (QString filter, output) {
        QAction *action = new QAction(filter, inputMenu);
        action->setCheckable(true);
        outputGroup->addAction( action );
        outputMenu->addAction( action );
    };
    outputGroup->actions().first()->setChecked( true );
    
    QStringList omits = QStringList() << "no" << "yes";
    QMenu *omitMenu = menu->addMenu("Omit zero");
    omitGroup = new QActionGroup(omitMenu);
    foreach (QString omit, omits) {
        QAction *action = new QAction(omit, omitMenu);
        action->setCheckable(true);
        omitGroup->addAction( action );
        omitMenu->addAction( action );
    };
    omitGroup->actions().first()->setChecked( true );
    
    
    QAction* menuAction = menu->addAction( "Generate output" );
    connect( menuAction, SIGNAL( triggered() ), this, SLOT( generateDotFileSelected() ) );

}


void 
CallgraphPlugin::cubeClosed()
{
}

/** set a version number, the plugin with the highest version number will be loaded */
void
CallgraphPlugin::version( int& major, int& minor, int& bugfix ) const
{
    major  = 1;
    minor  = 0;
    bugfix = 0;
}

QString
CallgraphPlugin::getHelpText() const
{
    return "Call tree based call graph generator.";
}

/** unique plugin name */
QString
CallgraphPlugin::name() const
{
    return "Call Graph Plugin";
}
